/*
mino Pico-FMeter2a
reziproke Frequenzmessung mit RP2040

Das Eingangssignal an 'FIN1_PIN' wird mit 'SystemCoreClock'/4 abgetastet und
die Zyklen mit dem X-Register gezaehlt. Bei einer fallenden Flanke wird der 
Inhalt des PIO0-SM0-X-Registers auf RXF0 ausgegeben und per DMA in die Variable 
F1_time_stamp geschrieben. Die Anzahl der Zeitstempel wird mit dem Register
DMA_TRANS_COUNT gezaehlt.

Nach Ablauf des vorgegebenen Messintervalls (DEF_F1_MESSZEIT) werden der
Zeitpunkt ('F1_time_stamp') und die Anzahl der Ereignisse (DMA_TRANS_COUNT)
ausgelesen und die Differenz zu den zuvor ausgelesen Werten gebildet. Die Zeit 
ist dabei synchron zu den Ereignissen. Die Frequenz ergibt sich als:
Ereignisse/Zeit * Abtastrate.

Die Abtastrate betraegt mit 'SystemCoreClock'/4 typisch 32,75 MHz. Bei dem
vorgegebenen Messintervall von 3 Messungen/s ist jede Messungen auf <= 1*10^-7 
(3 / 32,75*10^6) aufgeloest, was eine 7-stellige Aufloesung der Messwerte ergibt.
Die absolute Genauigkeit haengt von der Genauigkeit und Stabilitaet von 
'SystemCoreClock' ab.
Zur Korrektur dieser Frequenz dient zum einen die Variable 'f_ref_korrektur', 
die mit einem Trimmpoti am ADC0-Eingang eingestellt werden kann. 
Der korrigierte Wert wird als 'SM0_takt' als Referenzfrequenz verwendet.
Zum anderen wird das Ergebnis mit dem 'F1_offset' skaliert, welcher manuell fuer
internen (lokaler XO) oder externen (Fref) Takt manuell eingestellt werden kann
oder anhand eines ext. Referenzsignals 1 Hz / 10 kHz / 10 MHz ermittelt werden 
kann.

Zur feineren Aufloesung hoher Frequenzen wird bis 100 kHz Eingangsfrequenz
jede einzelne Periode gemessen und anhand einer linearen Regressionsberechnung
bewertet. Das ist die Grundfunktion.

Fuer Frequenzen >= 15 MHz muss ein externer Vorteiler verwendet werden, der hier
aus einem externen 4:1 Teiler und internen 8:1 (PWM_CH0) insgesamt 32:1 teilt. 
Dieser Vorteiler wird im MHz Bereich eingeschaltet. Dazu wird das Eingangssignal
von 'FIN1_PIN' auf 'PRE_I/O' umgeschaltet.
Fuer die Umschaltfrequenz wird mit 'PIO1_SM0' die grobe Frequenz 'F1H_grob' im
ms-Raster gemessen und in 'F1_messung()' dauernd bewertet.

Mit 'PIO1_SM1' wird zusaetzlich die Frequenz an 'F_REF' gemessen. Dies ist ein
weiterer Eingang fuer 0,01 Hz - 15 MHz dient aber auch als Referenzeingang fuer
extern angelegte Signale mit 1 Hz, 10 kHz oder 10 MHz.

Die uebrigen Funktionen sollten aus den Kommentaren erkennbar sein.


2024-02-23
Michael Nowak
http://www.mino-elektronik.de
Alle Angaben ohne Gewaehr !
*/

#include "System-Pico-Fmeter2.h"

#include <dma_req.h>
#include <intrinsics.h>
extern uint32_t *irq_tab;                          // fuer IRQ-Vektoren im RAM


// Umschaltschwellen fuer Vorteiler vorgeben >= 8 MHz ein, <= 4 MHz aus
#define F1H_FMIN          8000        // wegen 1 ms Intervall in kHz
#define F1_FMAX           4000        // dto.

#define DEF_STELLEN       8
#define DEF_FAKTOR        1
#define DEF_UPM_DIVISOR   1

#define F1H_VORTEILER     32          // gesamter Faktor fuer Skalierung
#define F1H_EXT_VORTEILER 4           // nur der ext. Vorteiler

enum {SM0 = 0, SM1, SM2, SM3} SMx_KANAL;
enum {CH0 = 0, CH1, CH2, CH3} DMA_CHx_Kanal;

#define FREF_FIL_LEN      8
#define FREF1             1e0         // 1 HZ
#define FREF1_LOW         (FREF1 - FREF1*5e-5)  // -50 ppm
#define FREF1_HIGH        (FREF1 + FREF1*5e-5)  // +50 ppm
#define FREF2             1e4         // 10 kHZ
#define FREF2_LOW         (FREF2 - FREF2*5e-5)  // -50 ppm
#define FREF2_HIGH        (FREF2 + FREF2*5e-5)  // +50 ppm
#define FREF3             1e6         // 1 MHz
#define FREF3_LOW         (FREF3 - FREF3*5e-5)  // -50 ppm
#define FREF3_HIGH        (FREF3 + FREF3*5e-5)  // +50 ppm
#define FREF4             1e7         // 10 MHz
#define FREF4_LOW         (FREF4 - FREF4*5e-5)  // -50 ppm
#define FREF4_HIGH        (FREF4 + FREF4*5e-5)  // +50 ppm

#define TEMP_STR_LEN      50          // fuer Ergebnis String
#define MAX_ERGEBNIS_STR  20          // Ausgabelaenge begrenzen

double rundung[] = {5e0, 5e-1,5e-2,5e-3,5e-4,5e-5,
                    5e-6,5e-7,5e-8,5e-9,5e-10,5e-11,
                    5e-12,5e-13,5e-14,5e-15};

char f_dim[][10][6]= {{" GHz"," MHz"," kHz"," Hz "," mHz"," sek"," ms "," us "," ns "," ps "},
                      {"E+9","E+6","E+3","E+0","E-3","E+0","E-3","E-6","E-9","E-12"}};  // Dimensionen zum Messwert
char u_dim[][5][6]=  {{" Grpm"," Mrpm"," krpm"," rpm"," mrpm"},
                      {"E+9","E+6","E+3","E+0","E-3"}};

enum anzeige {WANDEL_F=0, WANDEL_P, WANDEL_D};

uint16_t  gps_fil_cnt,          // Anzahl vorhandener Werte
          gps_fil_laenge = DEF_GPS_FILTER,
          ee_schreib_intervall, // gleiches intervall wie gps_fil_laenge
          gps_fil_index,        // Filter, Index auf Einzelwerte
          gps_totzeit;          // Einschaltverzoegerung abwarten

double    gps_fil[MAX_GPS_FILTER], // letzte Einzelwerte
          gps_fil_sum,
          ref_frequenz;         // akt. Summe der Einzelwerte

uint32_t  F1_time_stamp,        // Zeitpunkt von PIO0-SM0
          F1H_time_stamp,       // Zeitpunkt von PIO1-SM0
          Fref_time_stamp;      // Zeitpunkt von PIO1-SM1

// adjust missing counts of DMA
static uint32_t dma0_overflow, dma1_overflow, dma2_overflow;

uint16_t  led_count, gps_led_cnt, gps_fil_laenge,
          led_ein_zeit = TYP_LED_ZEIT;    // Blinkzeit

uint8_t   ausgabe_sperre = 1,             // keine LCD Datenausgage
          anzeige_funktion = F_UND_P,     // F+P, nur F, nur P, nur UPM oder auch aus -1
          update_LCD,                     // LCD neu beschriften
          nur_MHz,                        // festes Format x.xxx MHz
          auto_vorteiler_umschaltung,     // 1 = mit Neustart, 0 = schnell
          F1_reg_modus,
          anzeige_format = 0,             // bit0: Dimension oder Exponent; bit1: . oder , 
          F1_vorteiler_aktiv = 0,         // nicht verwendet
          F2_vorteiler_aktiv = 0,         // nicht verwendet
          gps_aktiv = 0,                  // nicht verwendet
          ser_ausgabe_kanal = F1;         // ser. Datenausgabe: F, P, UPM oder auch aus -1


double    F1_ergebnis;                    // in Hz
double    eff_zeit1;                      // in s
double volatile  
          F1_taktfrequenz;
uint32_t  F1_grob,                        // grobe Frequenzmessung mit 1 ms Intervall
          F1H_grob;                       // dto. fuer HF-Eingang
uint8_t   F1_F1H_auswahl;                 // Signaleingang: 0=ohne, 1= mit Vorteiler

uint32_t  F1_messzeit = DEF_F1_MESSZEIT,  // Sollintervall
          F2_messzeit = DEF_F2_MESSZEIT,  // Sollintervall
          F1_timeout = DEF_TIMEOUT,       // Abbruch bei fehlendem Signal
          F2_timeout = GPS_TIMEOUT,       // Abbruch bei fehlendem Signal
          F1_stellen = DEF_STELLEN,       // Vorgabewert
          F2_stellen = DEF_STELLEN,       // 
          F1_faktor  = DEF_FAKTOR,        // 1
          F2_faktor  = DEF_FAKTOR,        // 1
          F1_upm_divisor  = DEF_UPM_DIVISOR,  // ebenso
          F2_upm_divisor  = DEF_UPM_DIVISOR;  // ebenso
                    
volatile uint8_t
          F1_messwert_vorhanden;          // gueltige Messung
                                   
volatile uint32_t
          F1_mess_dauer,                  // in ms
          F1_timeout_counter;             // in ms
uint32_t  F1_start_zeit,                  // rel. Zeit: Beginn der Messung
          F1_end_zeit,                    // Zeitpunkt der Auswertung
          F1_mess_zeit;                   // genaue Zeit der Messung
                    
uint32_t  F1_mess_ereignisse,             // Impulse der Messung
          F1_start_ereignis,              // Impulse zu Beginn der Messung
          F1_end_ereignis;                // Impulse am Ende der Messung

int32_t   F1_offset = 0,
          Fintern_offset,
          Fextern_offset;

uint32_t  abgleich_zeit = MIN_GPS_FILTER;


double    eff_zeit2;                      // in s
double volatile     
          F_ref;                          // unkorrigierte Frequenz


// fuer Fref
double    Fref_ergebnis;                    // in Hz

volatile uint8_t
          Fref_messwert_vorhanden;          // 1 = gueltige Messung
                                   
volatile uint32_t
          Fref_mess_dauer,                  // in ms
          Fref_timeout_counter;             // in ms
uint32_t  Fref_start_zeit,                  // rel. Zeit: Beginn der Messung
          Fref_end_zeit,                    // Zeitpunkt der Auswertung
          Fref_mess_zeit;                   // genaue Zeit der Messung
                  
uint32_t  Fref_mess_ereignisse,             // Impulse der Messung
          Fref_start_ereignis,              // Impulse zu Beginn der Messung
          Fref_end_ereignis;                // Impulse am Ende der Messung

double    Fref_filter[FREF_FIL_LEN];
double    Fref_summe, 
          Fref_aktuell = 0.0,
          Fref_offset = 0.0;
uint32_t  Fref_index;

// Variablen fuer Regressionsberechnung
uint32_t  reg_n;                          // Anzahl Ereignisse bei Regression
volatile  uint32_t reg_ereignisse;        // und Endwert zur weiteren Auswertung
uint64_t  reg_xi, reg_yi;                 // summe fortlaufender Einzelwerte       
uint64_t  reg_sum_xi, reg_sum_yi;         // Summe aller Intervalle

double    reg_sum_xiyi,                   // Summe xi * yi
          reg_sum_xi2;                    // Summe xi
double    reg_xq, reg_yq;                 // Mittelwerte xi/n und yi/n

void get_F1_time_count(uint32_t *time, uint32_t *count);
void get_F1H_time_count(uint32_t *time, uint32_t *count);

// F1-offset fuer lokalen/externen Takt
void lade_F1_offset(void)
{
  if(ex_ref_aktiv == 0) {
    F1_offset = Fintern_offset;
  } else {
    F1_offset = Fextern_offset;
  } 
}

void schreibe_F1_offset(int offset)
{
  if(ex_ref_aktiv == 0) {
    Fintern_offset = offset;
  } else {
    Fextern_offset = offset;
  } 
}

// F1-offset ins EEPROM
void speicher_EE_offset(int offset)
{
    if(ex_ref_aktiv == 0) {
      ee_write_uint32(EE_ADR_Fintern_OFFSET, offset);
    } else {
      ee_write_uint32(EE_ADR_Fextern_OFFSET, offset);
    } 
}

void loesche_regression(void)
{
  reg_n = 0;                          // alle Werte der Regressionsberechnung
  reg_xi = reg_yi = 0;                // loeschen
  reg_sum_xi = reg_sum_yi = 0;
  reg_sum_xiyi = reg_sum_xi2 = 0;
  reg_xq = reg_yq = 0;
}

// ISR zu PWM6 mit 100 kHz
// jedes (Teil-)Intervall berechnen und weiterverarbeiten
void handle_F1_reg(void)
{
static uint8_t akt_eingang, discard_next;  // zum Umschalten des Vorteilers
// Test auf Vorteiler Umschaltung
  if(akt_eingang != F1_F1H_auswahl) {
    akt_eingang = F1_F1H_auswahl; // neuer Eingang
    if(akt_eingang)               // Signal nach Vorteiler verwenden
      PIO0->SM0_EXECCTRL = PIO0->SM0_EXECCTRL & ~(0x1F << PIO0_SM0_EXECCTRL_JMP_PIN_Pos) |
                           (FIN1H_PIN  << PIO0_SM0_EXECCTRL_JMP_PIN_Pos);  // Pin fuer JMP
    else                          // ohne Vorteiler messen
      PIO0->SM0_EXECCTRL = PIO0->SM0_EXECCTRL & ~(0x1F << PIO0_SM0_EXECCTRL_JMP_PIN_Pos) |
                           (FIN1_PIN  << PIO0_SM0_EXECCTRL_JMP_PIN_Pos);  // Pin fuer JMP
    discard_next = 3;             // drei nachfolgende Ergebisse verwerfen
    if(auto_vorteiler_umschaltung) {// 1 = Neustart der laufenden Messung erzwingen
      F1_mess_dauer =  F1_messzeit; // laufende Messung sofort abschliessen
      F1_messwert_vorhanden  = 0;   // und Ergebnis verwerfen
    }  
  }

// Zeitstempel fuer Einzelintervall lesen und auswerten
  get_F1_time_count(&F1_end_zeit, &F1_end_ereignis);
  F1_mess_zeit = F1_start_zeit-F1_end_zeit;
  F1_start_zeit = F1_end_zeit;
  F1_mess_ereignisse = (F1_start_ereignis-F1_end_ereignis);
  F1_start_ereignis = F1_end_ereignis;
  
  if(F1_mess_ereignisse) {              // nur gueltige Messwerte verwenden
    if(discard_next) {                  // Werte nach dem Umschalten verwerfen
      discard_next--;
    } else {
// Anzahl und Einzelwerte fuer die Regression verarbeiten 
      reg_n++;                          // anzahl der Einzelwerte
      if(akt_eingang) 
        reg_xi += F1_mess_ereignisse * F1H_VORTEILER;   // Einzelwerte addieren
      else
        reg_xi += F1_mess_ereignisse;   // Einzelwerte addieren
      reg_yi += F1_mess_zeit;           // dto
      reg_sum_xi += reg_xi;             // Summe aus Summe der Einzelwerte
      reg_sum_yi += reg_yi;             // dto
      reg_sum_xiyi += reg_xi*reg_yi;    // Summe der Produkte
      reg_sum_xi2  += reg_xi*reg_xi;    // Summe der Quadrate
    }  
  }  
}  

// Auswertung nach Ablauf der Messdauer starten  
double bewerte_F1_messung(void) 
{
static double reg_divisor;
double F1_temp;

  PWM->CH6_CSR =  0;                      // disable reg-timer
  if(reg_n) {
    reg_xq = (double)reg_sum_xi/(double)reg_n;
    reg_yq = (double)reg_sum_yi/(double)reg_n;
    reg_divisor = (reg_sum_xi2 - reg_n * (reg_xq * reg_xq));
    if(reg_n >= 2 && reg_divisor != 0.0) {  // Regression erst ab zwei Werten          
      F1_temp = SM0_takt / ((reg_sum_xiyi - reg_n * (reg_xq * reg_yq)) / reg_divisor);
    } else {                                // wenn nur ein Ereignis in reg_xi
      F1_temp = ((double)reg_xi/(double)reg_yi)*SM0_takt;  // Berechnung ohne Regression
    }
  } else F1_temp = 0.0;                     // kein Ergebnis moeglich 
// Endwerte fuer Debugger      
  reg_ereignisse = reg_n;
  if(reg_n) eff_zeit1 = reg_yi / SM0_takt;  // eff. Messzeit in s
  loesche_regression();
  PWM->CH6_CSR =  1;                      // enable reg-timer
  return(F1_temp);
}



uint8_t berechne_stellen(double zeit)
{
static double offset;
uint8_t stellen = MIN_STELLEN;	// minimal
  zeit *=10.0;                  // eine Stelle zusaetzlich
  if(zeit+offset >= 99.0) {
    stellen = 10;
    offset = 10.0;
  } else if(zeit+offset >= 29) {
    stellen = 9;
    offset = 3.0;
  } else if(zeit+offset >= 2.9) {
    stellen = 8;
    offset = 0.3;
  } else if(zeit+offset >= 2.9e-1) {
    stellen = 7;
    offset = 0.03;
  } else if(zeit+offset >= 2.9e-2) {
    stellen = 6;
    offset = 0.003;
  } else if(zeit+offset >= 2.9e-3) {
    stellen = 5;
    offset = 0.0003;
  } else offset = 0;
//  if(stellen < min_stellen) stellen = min_stellen;
  return(stellen);
}

// bedingte Anzeige von Messwerten
void messwert_lcd1(char *s)
{
  if(!ausgabe_sperre) lcd_messwert_zeile1(s);
}  

void messwert_lcd2(char *s)
{
  if(!ausgabe_sperre) lcd_messwert_zeile2(s);
} 

// Routine zur Wandlung und Ausgabe eines Messwertes
void wandel_Fx(char *header, double x, char zeige_periode, char *s, uint8_t stellen)    // Anzeige von F, P oder UPM
{
int i,j,dez_punkt,dimension;
char *p;
  if(zeige_periode == WANDEL_F && nur_MHz)
  {
    if(x >= 1e6) stellen--;
    if(x >= 1e7) stellen--;
    if(x >= 1e8) stellen--;
    if(x >= 1e9) stellen--;
    sprintf(s,"%.*f MHz",stellen,x*1e-6);   // festes Format bei "nur MHz"
    if(*s == '0') *s = ' ';
    return;
  }    

  dez_punkt = 0;
  dimension = 3;
  p = header;
  while(*p) *s++ = *p++;                    // "P = " fuer Periodendauer
  if(x >= 0.001) {                          // 1mHz ist Untergrenze
    if(zeige_periode == WANDEL_P) {
      x = 1/x;                              // Kehrwert bilden
      dimension=5;                          // und in Sekunden als Dimension
    }
    while(x<1.0) {x*=1000.0;dimension++;}   // in den Hz-Bereich bringen
    while(x>=1000.0) {x*=0.001;dimension--;}
    while(x >= 10.0) {
      x *= 0.1;
      dez_punkt++;
    }
    x += rundung[stellen];                  // runden
    if(x >= 10.0) {			                    // Ueberlauf bei Rundung
      x *= 0.1;                             // korrigieren
      dez_punkt++;
      if(dez_punkt > 2) {		                // Ueberlauf der Dimension
        dimension--;
        dez_punkt = 0;
      }
    }
  } else {
    x = 0.0; dez_punkt=0; dimension=3;      // 0.00000 Hz ausgeben
  }
  
  for(i=0;i<stellen;i++) {
    j = (uint8_t)x;
    *s++ = (j+'0');
    if(i == dez_punkt) {
      if(anzeige_format & 2)
        *s++ = (',');
      else
        *s++ = ('.');
    }  
    x -= j;
    x *= 10;
  }

  if(zeige_periode == WANDEL_D)
    p = u_dim[anzeige_format&1][dimension];
  else
    p = f_dim[anzeige_format&1][dimension];
  while(*p) *s++ = *p++;                    // Dimension anhaengen
  *s++ = 0;                                 // string abschliessen
  *s = 0;                                   // string abschliessen
  s[MAX_ERGEBNIS_STR] = 0;                  // zur Sicherheit begrenzen
}

void Fx_str_mittig(char *header, double x, char zeige_periode, char *s, uint8_t stellen)    // Anzeige von F, P oder UPM
{
char p[TEMP_STR_LEN];
int i,j;
  *p = 0;
  for(i = 0; i < MAX_ERGEBNIS_STR; i++)     // Ergebnisstring loeschen
    s[i] = ' ';
  s[MAX_ERGEBNIS_STR] = 0;                  // Ende von s
  
  wandel_Fx(header,x,zeige_periode,p,stellen);// wandeln
  i = strlen(p);
  i = lcd_breite - i;                       // auf 16 Zeichen/Zeile einmitten
  if(i < 0) i = 0;                          // Unterlauf verhindern
  j = i/2;
  i = 0;
  while(p[i]) s[j++] = p[i++];              // in den Ergebnisstring schreiben
  s[lcd_breite] = 0;                        // Laenge sicher begrenzen
} 

/*
 PIO0-SM0:
 Zeitzaehlung im x-Register mit SystemCoreClock/PIO_TIME_DIV (=4)
 und Ausgabe des Wertes bei negativer Flanke des Eingangssignals
*/
void init_PIO0_SM0(void)
{
// PIO0 aktivieren  
  RESETS_CLR->RESET = RESETS_RESET_pio0_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_pio0_Msk));

// FIN1_PIN fuer SM0 initialisieren
  IO_BANK0->GPIO18_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;  // GPIO-Eingang
  PADS_BANK0->GPIO18 = PADS_BANK0_GPIO18_IE_Msk |             // pullup + schmitttrigger
                      PADS_BANK0_GPIO18_PUE_Msk | 
                      PADS_BANK0_GPIO18_SCHMITT_Msk;
  
  PIO0->SM0_EXECCTRL = (31 << PIO0_SM0_EXECCTRL_WRAP_TOP_Pos) | 
                        (0 << PIO0_SM0_EXECCTRL_WRAP_BOTTOM_Pos) |
                        (FIN1_PIN  << PIO0_SM0_EXECCTRL_JMP_PIN_Pos);  // Pin fuer JMP
// ADR0  
  PIO0->INSTR_MEM0 = 0x0241;      // jmp x-- -> ADR1, 2 x delay
// ADR1
  PIO0->INSTR_MEM1 = 0x00c0;      // jmp FIN1_PIN -> ADR0 warten bis '0'
// input-pin 1 -> 0
  PIO0->INSTR_MEM2 = 0x4020;      // mov x->isr, x-Wert ausgeben
// ADR3
  PIO0->INSTR_MEM3 = 0x0044;      // jmp x-- -> ADR4
// ADR4  
  PIO0->INSTR_MEM4 = 0x00c1;      // jmp FIN1_PIN 0-> 1 dann ADR1
  PIO0->INSTR_MEM5 = 0x0103;      // jmp ADR3, 1 x delay
  
  PIO0->SM0_CLKDIV = 1 << PIO0_SM0_CLKDIV_INT_Pos |
                     0 << PIO0_SM0_CLKDIV_FRAC_Pos;
  PIO0_SET->SM0_SHIFTCTRL = PIO0_SM0_SHIFTCTRL_FJOIN_RX_Msk | // FIFO = 8
                            PIO0_SM0_SHIFTCTRL_AUTOPUSH_Msk;  // autopush
  PIO0_SET->CTRL = (BIT(SM0) << (PIO0_CTRL_SM_ENABLE_Pos));   // starten
}

// F1H-Messung
void init_PIO1_SM0(void)
{
// enable PIO1  
  RESETS_CLR->RESET = RESETS_RESET_pio1_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_pio1_Msk));
  
// PRE_I/O fuer PIO1_SM0 initialisieren
  IO_BANK0->GPIO16_CTRL = IO_BANK0_GPIO16_CTRL_FUNCSEL_pwm_a_0; // PWM-Ausgang Vorteiler
  SIO->GPIO_OE_SET = BIT(FIN1H_PIN);
// F_PRE auf PWM0B-input  
  IO_BANK0->GPIO17_CTRL = IO_BANK0_GPIO17_CTRL_FUNCSEL_pwm_b_0; // PWM0-Eingang
  PADS_BANK0->GPIO17 = PADS_BANK0_GPIO17_IE_Msk |               // pullup + schmitttrigger
                      PADS_BANK0_GPIO17_PUE_Msk | 
                      PADS_BANK0_GPIO17_SCHMITT_Msk;
// interner Vorteiler mit PWM0-Kanal  
  PWM->CH0_TOP = F1H_VORTEILER/F1H_EXT_VORTEILER-1;             // interner Wert fuer 8:1
  PWM->CH0_CC  = F1H_VORTEILER/F1H_EXT_VORTEILER/2;             // 50% Tastverhaeltnis
  PWM_SET->CH0_CSR = PWM_CH0_CSR_DIVMODE_Msk | PWM_CH0_CSR_EN_Msk;  // fallende Flanke + enable
 
// Eingangspin SM0 festlegen  
  PIO1->SM0_EXECCTRL = (31 << PIO0_SM0_EXECCTRL_WRAP_TOP_Pos) | 
                        (0 << PIO0_SM0_EXECCTRL_WRAP_BOTTOM_Pos) |
                        (FIN1H_PIN  << PIO0_SM0_EXECCTRL_JMP_PIN_Pos);  // Pin fuer JMP
  PIO1_SET->SM0_SHIFTCTRL = PIO0_SM0_SHIFTCTRL_FJOIN_RX_Msk | // FIFO = 8
                            PIO0_SM0_SHIFTCTRL_AUTOPUSH_Msk;  // autopush
// ADR0  
  PIO1->INSTR_MEM0 = 0x0241;      // jmp x-- -> ADR1, 2 x delay
// ADR1
  PIO1->INSTR_MEM1 = 0x00c0;      // jmp FIN1_PIN -> ADR0 warten bis '0'
// input-pin 1 -> 0
  PIO1->INSTR_MEM2 = 0x4020;      // mov x->isr, x-Wert ausgeben
// ADR3
  PIO1->INSTR_MEM3 = 0x0044;      // jmp x-- -> ADR4
// ADR4  
  PIO1->INSTR_MEM4 = 0x00c1;      // jmp FIN1_PIN 0-> 1 dann ADR1
  PIO1->INSTR_MEM5 = 0x0103;      // jmp ADR3, 1 x delay

  PIO1->SM0_CLKDIV = 1 << PIO0_SM0_CLKDIV_INT_Pos |
                     0 << PIO0_SM0_CLKDIV_FRAC_Pos;
  PIO1_SET->CTRL = (BIT(SM0) << (PIO0_CTRL_SM_ENABLE_Pos));   // PIO1_SM0 starten
}

void init_PIO1_SM1(void)
{
// PIO1 aktivieren  
// Programm ab ADR0 wie bei PIO1_SM0  
  RESETS_CLR->RESET = RESETS_RESET_pio1_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_pio1_Msk));

// FREF_PIN fuer SM0 initialisieren
  IO_BANK0->GPIO19_CTRL = IO_BANK0_GPIO19_CTRL_FUNCSEL_sio_19;// GPIO-Eingang
  PADS_BANK0->GPIO19 = PADS_BANK0_GPIO19_IE_Msk |             // pullup + schmitttrigger
                      PADS_BANK0_GPIO19_PUE_Msk | 
                      PADS_BANK0_GPIO19_SCHMITT_Msk;

// Eingangspin SM1 festlegen  
  PIO1->SM1_EXECCTRL = (31 << PIO0_SM1_EXECCTRL_WRAP_TOP_Pos) | 
                        (0 << PIO0_SM1_EXECCTRL_WRAP_BOTTOM_Pos) |
                        (FREF_PIN  << PIO0_SM1_EXECCTRL_JMP_PIN_Pos);  // Pin fuer JMP
  PIO1_SET->SM1_SHIFTCTRL = PIO0_SM1_SHIFTCTRL_FJOIN_RX_Msk | // FIFO = 8
                            PIO0_SM1_SHIFTCTRL_AUTOPUSH_Msk;  // autopush
  
  PIO1->SM1_CLKDIV = 1 << PIO0_SM0_CLKDIV_INT_Pos |
                     0 << PIO0_SM0_CLKDIV_FRAC_Pos;
  PIO1_SET->CTRL = (BIT(SM1) << (PIO0_CTRL_SM_ENABLE_Pos));   // PIO1_SM1 starten
}

// DMA-Counter nachtriggern wenn abgelaufen
void DMA_IRQ_0_IRQHandler(void)
{
  if(DMA->INTR & BIT(CH0)) {             // DMA-CH0 count == 0?
    DMA->INTS0 = BIT(CH0);               // clear flag
    DMA_SET->CH0_CTRL_TRIG = DMA_CH0_CTRL_TRIG_EN_Msk;    // restart DMA-CH0
		dma0_overflow++;
	}	
  if(DMA->INTR & BIT(CH1)) {             // DMA-CH1 count == 0?
    DMA->INTS0 = BIT(CH1);               // clear flag
    DMA_SET->CH1_CTRL_TRIG = DMA_CH1_CTRL_TRIG_EN_Msk;    // restart DMA-CH1
		dma1_overflow++;
  }
  if(DMA->INTR & BIT(CH2)) {             // DMA-CH2 count == 0?
    DMA->INTS0 = BIT(CH2);               // clear flag
    DMA_SET->CH2_CTRL_TRIG = DMA_CH2_CTRL_TRIG_EN_Msk;    // restart DMA-CH2
		dma2_overflow++;
  }
  __DSB();
}

// Zeitstempel erzeugen:
// DMA-Wert = Zeitpunkt, DMA-TRANS-COUNT = Ereignisse
void init_dma_pio0_1(void)
{
  RESETS_CLR->RESET = RESETS_RESET_dma_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_dma_Msk));

  irq_tab[DMA_IRQ_0_IRQn] = (uint32_t)DMA_IRQ_0_IRQHandler;  // Vektor im RAM
  NVIC_EnableIRQ(DMA_IRQ_0_IRQn);
  NVIC_SetPriority(DMA_IRQ_0_IRQn,2);
// PIO0_SM0  
  DMA->CH0_READ_ADDR = (uint32_t) &PIO0->RXF0;
  DMA->CH0_WRITE_ADDR = (uint32_t) &F1_time_stamp;
  DMA->CH0_TRANS_COUNT = -1;
  DMA->CH0_CTRL_TRIG =  DREQ_PIO0_RX0 << DMA_CH0_CTRL_TRIG_TREQ_SEL_Pos |     // PIO0_RXF0
                        2 << DMA_CH0_CTRL_TRIG_DATA_SIZE_Pos |  // word
                        DMA_CH0_CTRL_TRIG_EN_Msk;               // enable
// PIO1_SM0  
  DMA->CH1_READ_ADDR = (uint32_t) &PIO1->RXF0;
  DMA->CH1_WRITE_ADDR = (uint32_t) &F1H_time_stamp;
  DMA->CH1_TRANS_COUNT = -1;
  DMA->CH1_CTRL_TRIG =  DREQ_PIO1_RX0 << DMA_CH1_CTRL_TRIG_TREQ_SEL_Pos |     // PIO1_RXF0
                        2 << DMA_CH0_CTRL_TRIG_DATA_SIZE_Pos |  // word
                        DMA_CH1_CTRL_TRIG_EN_Msk;               // enable
// PIO1_SM1  
  DMA->CH2_READ_ADDR = (uint32_t) &PIO1->RXF1;
  DMA->CH2_WRITE_ADDR = (uint32_t) &Fref_time_stamp;
  DMA->CH2_TRANS_COUNT = -1;
  DMA->CH2_CTRL_TRIG =  DREQ_PIO1_RX1 << DMA_CH2_CTRL_TRIG_TREQ_SEL_Pos |     // PIO1_RXF1
                        2 << DMA_CH0_CTRL_TRIG_DATA_SIZE_Pos |  // word
                        DMA_CH0_CTRL_TRIG_EN_Msk;               // enable
  
  DMA->INTS0 = BIT(CH0) | BIT(CH1) | BIT(CH2);           // clear flag CH0+CH1+CH2
  DMA_SET->INTE0 = BIT(CH0) | BIT(CH1) | BIT(CH2);       // enable CH0+CH1+CH2-isr
}


// 1 kHz Aufrufe fuer Messzeiten F1_TIMER (PWM7)
// bzw. 100 kHz fuer Regressionsberechnung REG_TIMER (PWM6)
void PWM_IRQ_WRAP_IRQHandler(void)
{
static uint32_t F1_time_stamp_old,                // timeout testen
                F1H_time_stamp_old,
                Fref_time_stamp_old;
static uint32_t F1H_end_ereignis,
                F1H_end_zeit,                     // dummy
                F1_grob_ereignis,                 // letzter wert
                F1H_grob_ereignis;
  
  if(PWM->INTS & BIT(REG_TIMER)) {                  // PWM_CH6?
    PWM->INTR = BIT(REG_TIMER);                     // clear flag
    handle_F1_reg();
  }  
  if(PWM->INTS & BIT(F1_TIMER)) {                   // PWM_CH7?
    PWM->INTR = BIT(F1_TIMER);                      // clear flag
    
// PIO0_SM0    
    F1_mess_dauer++;
    F1_timeout_counter++;
    F1_grob = (F1_grob_ereignis-F1_end_ereignis); // ueberschlaegige Frequenzbestimmung im 1 ms Intervall
    F1_grob_ereignis = F1_end_ereignis;           // fuer naechstes intervall
    if(F1_time_stamp != F1_time_stamp_old) {      // new count?
      F1_time_stamp_old = F1_time_stamp;          // yes
      F1_timeout_counter = 0;                     // reset timeout
    }
// PIO1_SM0
    get_F1H_time_count(&F1H_end_zeit, &F1H_end_ereignis);
    F1H_grob = (F1H_grob_ereignis-F1H_end_ereignis)*F1H_VORTEILER;  // ueberschlaegige Frequenzbestimmung im 1 ms Intervall
    F1H_grob_ereignis = F1H_end_ereignis;         // fuer naechstes intervall
    if(F1H_time_stamp != F1H_time_stamp_old) {    // new count?
      F1H_time_stamp_old = F1H_time_stamp;        // yes
      F1_timeout_counter = 0;                     // reset timeout
    }
// PIO1_SM1    
    Fref_mess_dauer++;
    Fref_timeout_counter++;
    if(Fref_time_stamp != Fref_time_stamp_old) {  // new count?
      Fref_time_stamp_old = Fref_time_stamp;      // yes
      Fref_timeout_counter = 0;                   // reset timeout
    }
    dekodiere_tasten();
// fertig-LED und 1pps-LED blinken lassen
    if(led_count) {                               // ready-LEDs
      SIO->GPIO_OUT_SET = BIT(LOCAL_LED) | BIT(EXT_LED);  // on
      led_count--;
    } else {
      SIO->GPIO_OUT_CLR = BIT(LOCAL_LED) | BIT(EXT_LED);  // off
    }  
    if(gps_led_cnt) {
      gps_led_cnt--;
      SIO->GPIO_OUT_SET = BIT(GPS_LED);
    } else {
      SIO->GPIO_OUT_CLR = BIT(GPS_LED);
    }    
  }
  __DSB();
}

// ms-Raster fuer Messablauf erzeugen
// 100 kHz per PWM6 fuer Regressionsberechnung starten
void init_F1_timer(void)
{
  RESETS_CLR->RESET = RESETS_RESET_pwm_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_pwm_Msk));
  IO_BANK0->GPIO14_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;  // LED-Ausgang
  IO_BANK0->GPIO15_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;  // LED-Ausgang
  IO_BANK0->GPIO25_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;  // LED-Ausgang
  SIO->GPIO_OE_SET = BIT(LOCAL_LED); 
  SIO->GPIO_OE_SET = BIT(EXT_LED); 
  SIO->GPIO_OE_SET = BIT(GPS_LED); 

// externe Taster als Eingaenge mit Pullup  
  PADS_BANK0->GPIO10 = PADS_BANK0_GPIO19_IE_Msk |   // pullup + schmitttrigger
                      PADS_BANK0_GPIO19_PUE_Msk | 
                      PADS_BANK0_GPIO19_SCHMITT_Msk;
  PADS_BANK0->GPIO11 = PADS_BANK0_GPIO19_IE_Msk |   // pullup + schmitttrigger
                      PADS_BANK0_GPIO19_PUE_Msk | 
                      PADS_BANK0_GPIO19_SCHMITT_Msk;
  PADS_BANK0->GPIO12 = PADS_BANK0_GPIO19_IE_Msk |   // pullup + schmitttrigger
                      PADS_BANK0_GPIO19_PUE_Msk | 
                      PADS_BANK0_GPIO19_SCHMITT_Msk;
  
  
  PWM->CH7_DIV = 100 << PWM_CH0_DIV_INT_Pos;      // Vorteiler / 100
  PWM->CH7_TOP = SystemCoreClock/100/1000-1;      // 1 kHz
  PWM->CH7_CSR =  1;                              // enable ms-timer
  
  PWM->CH6_DIV = 100 << PWM_CH0_DIV_INT_Pos;      // Vorteiler / 100
  PWM->CH6_TOP = SystemCoreClock/100/100000-1;    // 100 kHz
  PWM->CH6_CSR =  1;                              // enable reg-timer
  
  irq_tab[PWM_IRQ_WRAP_IRQn] = (uint32_t)PWM_IRQ_WRAP_IRQHandler;  // Vektor im RAM
  NVIC_EnableIRQ(PWM_IRQ_WRAP_IRQn);
  NVIC_SetPriority(PWM_IRQ_WRAP_IRQn,3);
  
  PWM_SET->INTE = BIT(F1_TIMER) | BIT(REG_TIMER); // beide Timer per iSR                                // enable ISR
}


// Zeitstempel F1 lesen
void get_F1_time_count(uint32_t *time, uint32_t *count)
{
uint32_t temp1, temp2;  
__disable_interrupt();
  DMA_CLR->CH0_CTRL_TRIG = DMA_CH0_CTRL_TRIG_EN_Msk;  // stop DMA
  __DSB();  // stop unbedingt abwarten
  temp1 =   DMA->CH0_AL1_TRANS_COUNT_TRIG + dma0_overflow;            // count
  temp2 = F1_time_stamp;                              // get time stamp
  DMA_SET->CH0_CTRL_TRIG = DMA_CH0_CTRL_TRIG_EN_Msk;  // continue DMA
__enable_interrupt();
  *count = temp1;
  *time  = temp2;
}  

// Zeitstempel F1H lesen
void get_F1H_time_count(uint32_t *time, uint32_t *count)
{
uint32_t temp1, temp2;  
__disable_interrupt();
  DMA_CLR->CH1_CTRL_TRIG = DMA_CH1_CTRL_TRIG_EN_Msk;  // stop DMA
  __DSB();  // stop unbedingt abwarten
  temp1 =   DMA->CH1_AL1_TRANS_COUNT_TRIG + dma1_overflow;            // count
  temp2 = F1H_time_stamp;                             // get time stamp
  DMA_SET->CH1_CTRL_TRIG = DMA_CH1_CTRL_TRIG_EN_Msk;  // continue DMA
__enable_interrupt();
  *count = temp1;
  *time  = temp2;
}

// Zeitstempel Fref lesen
void get_Fref_time_count(uint32_t *time, uint32_t *count)
{
uint32_t temp1, temp2;  
__disable_interrupt();
  DMA_CLR->CH2_CTRL_TRIG = DMA_CH2_CTRL_TRIG_EN_Msk;  // stop DMA
  __DSB();  // stop unbedingt abwarten
  temp1 =   DMA->CH2_AL1_TRANS_COUNT_TRIG + dma2_overflow;            // count
  temp2 = Fref_time_stamp;                            // get time stamp
  DMA_SET->CH2_CTRL_TRIG = DMA_CH2_CTRL_TRIG_EN_Msk;  // continue DMA
__enable_interrupt();
  *count = temp1;
  *time  = temp2;
}  
     


void F1_messung(uint8_t init_timer)
{
static uint8_t  init = 0, alte_funktion = 255;
uint8_t eff_stellen, temp;
static char p_str[TEMP_STR_LEN];
static char s[TEMP_STR_LEN];
const char *str0[] = {
                "    Signal?     ",
                "    signal?     "};
const char *stry[] = {
                "    Signal F?    ",
                "    signal F?    "};
const char *strz[] = {
                "    Signal P?    ",
                "    signal P?    "};
const char *str1[] = {
                "    Frequenz    ",
                "   frequency    "};
const char *str2[] = {
                "    Periode    ",
                "    periode    "};
const char *str3[] = {
                "    Drehzahl    ",
                " rotation/min.  "};
  if(init_timer) { 
    init = 0;                           // Anzeige + Messung initialisieren
    init_F1_timer();                    // 1 ms Takt und 100 kHz Regression
    init_dma_pio0_1();                  // time stamps DMA
    init_PIO0_SM0();                    // PIO0 SM0 starten
    init_PIO1_SM0();                    // PIO1 SM0 fuer F1H starten
    if(!ex_ref_aktiv)
      init_PIO1_SM1();                  // PIO1 SM1 nur bei lokalem Takt aktivieren
    lade_F1_offset();                   // beim Start festlegen
  }
   
  if(!init) {                           // bei Neustart
    F1_mess_dauer =  F1_messzeit;       // 1. Messung sofort abschliessen
    F1_messwert_vorhanden  = 0;         // und 1. Ergebnis verwerfen
    F1_timeout_counter = 0; 
    init = 1;
    update_LCD = 1;
    init_stat();
  }
 
  if(!ausgabe_sperre && (anzeige_funktion != alte_funktion || update_LCD)) {  
    update_LCD = 0;
    if(anzeige_funktion >= MAX_FUNKTION) 
      anzeige_funktion = F_UND_P;       // sinnvollen Wert einstellen
    alte_funktion = anzeige_funktion;
    loesche_lcd_zeile12();
    switch(anzeige_funktion) {
      case F_UND_P:
        lcd_zeile1((char *)stry[sprache]);
        lcd_zeile2((char *)strz[sprache]);
        break;
      case FREQUENZ:
        lcd_zeile1((char *)str1[sprache]);
        lcd_zeile2((char *)str0[sprache]);
        break;
      case PERIODE:
        lcd_zeile1((char *)str2[sprache]);
        lcd_zeile2((char *)str0[sprache]);
        break;
      case DREHZAHL:
        lcd_zeile1((char *)str3[sprache]);
        lcd_zeile2((char *)str0[sprache]);
        break;
      case FMEAN_UND_N:
        lcd_zeile1("  Fmean: n=     ");
        lcd_zeile2((char *)str0[sprache]);
        break;
      case FMAX_FMIN:
        lcd_zeile1("   Fmax-Fmin    ");
        lcd_zeile2((char *)str0[sprache]);
        break;
      case SDEV_UND_N:
        lcd_zeile1("  S-dev: n =    ");
        lcd_zeile2((char *)str0[sprache]);
        break;
    }
  }
  
  if(F1_timeout_counter >= F1_timeout) {      // bei fehlenden Eingangsimpulsen
    init = 0;                                 // neu initialisieren
  }  
// grobe bewertung von Fin zur Auswahl des passenden Kanals
  if(F1H_grob > F1_FMAX) F1_F1H_auswahl = 1;        // F1H nehmen fuer f > 8 MHz
  else if(F1H_grob < F1H_FMIN) F1_F1H_auswahl = 0;  // F1 fuer f < 4 MHz 
    
  if(F1_mess_dauer >= F1_messzeit) {          // wenn Mindestzeit erreicht
    F1_ergebnis = bewerte_F1_messung();       // F1_ergebnis und eff_zeit1 berechnen
    if(F1_ergebnis > 0.0) {
// lokale Taktfrequenz SM0_takt ist mit Poti korrigiert
// digitalen Abgleich mit F1_offset durchfuehren   
      F1_ergebnis += F1_ergebnis*(F1_offset*F_OFFSET_SCHRITT);
      F1_mess_dauer = 0;                      // neue Messung beginnen
      if(F1_messwert_vorhanden) { 
        eff_stellen = F1_stellen;
        if(!eff_stellen) eff_stellen = berechne_stellen(eff_zeit1);
        if(F1_vorteiler_aktiv)
          F1_ergebnis *= (double)F1_faktor;   // und ggf. skalieren
        stat_new(F1_ergebnis, eff_stellen);  
        led_count = led_ein_zeit;
// serielle Datenausgabe zuerst        
        if(ser_ausgabe_kanal == P1) {
          wandel_Fx("",F1_ergebnis, WANDEL_P, s, eff_stellen);
          sende_str1(s);
        }
        if(ser_ausgabe_kanal == UPM1) {
          wandel_Fx("",F1_ergebnis*60.0/F1_upm_divisor, WANDEL_D, s, eff_stellen);
          sende_str1(s);
        }
        wandel_Fx("",F1_ergebnis, WANDEL_F, s, eff_stellen);
        if(ser_ausgabe_kanal == F1) 
          sende_str1(s);
// danach Anzeige auf LCD        
        switch(anzeige_funktion) {
          case F_UND_P:
            Fx_str_mittig("F: ",F1_ergebnis, WANDEL_F, p_str, eff_stellen);
            messwert_lcd1(p_str);
            Fx_str_mittig("P: ",F1_ergebnis, WANDEL_P, p_str, eff_stellen);
            messwert_lcd2(p_str);
            break;
          case FREQUENZ:
            Fx_str_mittig("",F1_ergebnis, WANDEL_F, p_str, eff_stellen);
            messwert_lcd2(p_str);
            break;
          case PERIODE:
            Fx_str_mittig("",F1_ergebnis, WANDEL_P, p_str, eff_stellen);
            messwert_lcd2(p_str);
            break;
          case DREHZAHL:
            Fx_str_mittig("",F1_ergebnis*60.0/F1_upm_divisor, WANDEL_D, s, eff_stellen);
            messwert_lcd2(s);
            break;
          case FMEAN_UND_N:
            strcpy(s,"Fmean: n=     ");
            long2a(stat_n, &s[10], 0);
            LCD_str_einmitten(s, lcd_breite);
            messwert_lcd1(s);
            Fx_str_mittig("\001 ",stat_mean, WANDEL_F, s, eff_stellen);
            messwert_lcd2(s);
            break;
          case FMAX_FMIN:
            Fx_str_mittig("\002 ",stat_max, WANDEL_F, p_str, eff_stellen);
            messwert_lcd1(p_str);
            Fx_str_mittig("\003 ",stat_min, WANDEL_F, p_str, eff_stellen);
            messwert_lcd2(p_str);
            break;
          case SDEV_UND_N:
            strcpy(s,"S-dev: n=     ");
            long2a(stat_n, &s[10], 0);
            LCD_str_einmitten(s, lcd_breite);
            messwert_lcd1(s);
            temp = anzeige_format;
            anzeige_format |= 1;
            if(stat_n >= 1) 
              Fx_str_mittig("\345 ",stat_sdev(), WANDEL_F, s, eff_stellen);
            anzeige_format = temp;
            messwert_lcd2(s);
            break;
        }
      } else F1_messwert_vorhanden  = 1;    // sperre wieder aufheben
    }
  }   
}  


void loesche_gps_filter(void)
{
  gps_fil_laenge = abgleich_zeit;
  for(gps_fil_index = 0; gps_fil_index < gps_fil_laenge; gps_fil_index++)
    gps_fil[gps_fil_index] = 0l;
  gps_fil_index = 0;
  gps_fil_cnt = 0;
  gps_fil_sum = 0.0;
  gps_led_cnt = 0;                          // GPS-Signal fehlt: LED ausschalten
  ee_schreib_intervall = gps_fil_laenge;    // neu warten
  gps_totzeit = GPS_TOTZEIT;
}

void bewerte_gps_signal(double frequenz_1Hz) {    // auf 1 Hz normiert
uint16_t  led_temp; 
double f;

  gps_fil_sum -= gps_fil[gps_fil_index];    // letzen Wert abziehen
  gps_fil_sum += frequenz_1Hz;              // neuen Wert addieren
  gps_fil[gps_fil_index++] = frequenz_1Hz;  // und ablegen
  if(gps_fil_index >= gps_fil_laenge)
    gps_fil_index = 0;
  gps_fil_cnt++;                            // mitzaehlen, ob Filter 'gefuellt'
  if(gps_fil_cnt >= gps_fil_laenge) {       // Filter ist 'eingeschwungen'
    gps_fil_cnt = gps_fil_laenge;           // nicht weiter zaehlen
    ref_frequenz = (gps_fil_sum / gps_fil_laenge) * F_ref; // wieder skalieren
    gps_led_cnt = GPS_TIMEOUT;              // Dauerlicht durch Nachtriggern
    F1_taktfrequenz = ref_frequenz;         // jede Sekunde neu anpassen
// Warnung wg. volatile vermeiden
    f = F_ref - ref_frequenz;               // auf 0,1 ppb/Schritt
    F1_offset = (int)(f / F_OFFSET_SCHRITT/F_ref);// skalieren
    ee_schreib_intervall++;
    if(ee_schreib_intervall >= gps_fil_laenge) {
      ee_schreib_intervall = 0;             // neu zaehlen
      schreibe_F1_offset(F1_offset);        // nach Fintern_offset
      speicher_EE_offset(F1_offset);        // zyklisch letzten Wert speichern                                      
    }
  } else {
    if(gps_fil_laenge >= GPS_LED_STUFEN) {  // Abstufung, um zu kurze Zeiten zu verhindern
      led_temp = (gps_fil_cnt/(gps_fil_laenge/GPS_LED_STUFEN))*(gps_fil_laenge/GPS_LED_STUFEN);
      if(led_temp <= (gps_fil_laenge/GPS_LED_STUFEN)/2) 
        led_temp = (gps_fil_laenge/GPS_LED_STUFEN)/2+1;  // minimal 1
    } else led_temp = gps_fil_cnt;                  // Dauerlicht, wenn Filter gefuellt ist
    gps_led_cnt = (1000*led_temp)/gps_fil_laenge;   // Blinkzeit in ms, wenn Filter noch nicht voll ist
  }
}

void Fgps_messung(void)
{
static uint8_t  init = 0, gps_laeuft = 0;
uint32_t eff_stellen;
static char s[TEMP_STR_LEN];
double F_temp, Fref_temp;
const char *str0[] = {
                "  Ref. Signal   ",
                "  ref. signal   "};
const char *str1[] = {
                " Frequenz F-Ref ",
                "frequency F-ref "};
const char *str2[] = {
                "    Signal?     ",
                "    signal?     "};

    if(!init) {
      Fref_mess_dauer =  F2_messzeit;         // und 1. Messung sofort abschliessen
      Fref_messwert_vorhanden  = 0;             // dabei 1. Ergebnis verwerfen
      Fref_timeout_counter = 0;
      if(gps_aktiv)
        lcd_zeile3((char *)str0[sprache]); 
      else
        lcd_zeile3((char *)str1[sprache]); 
      lcd_zeile4((char *)str2[sprache]);
      F_ref = (SystemCoreClock/PIO_TIME_DIV);
      init = 1;
    } 
  
  if(Fref_timeout_counter >= F2_timeout) {    // bei fehlenden Eingangsimpulsen
    init = 0;                                   // neu initialisieren
  }  
  if(Fref_mess_dauer >= F2_messzeit) {        // wenn Mindestzeit erreicht
    get_Fref_time_count(&Fref_end_zeit, &Fref_end_ereignis);
// Fref
    Fref_mess_zeit = Fref_start_zeit-Fref_end_zeit;
    Fref_start_zeit = Fref_end_zeit;
    Fref_mess_ereignisse = Fref_start_ereignis-Fref_end_ereignis;
    Fref_start_ereignis = Fref_end_ereignis;

    Fref_mess_dauer = 0;                        // und Messzeit abwarten
    if(Fref_mess_ereignisse) {
      if(Fref_messwert_vorhanden) {             // Ergebnis berechnen und anzeigen/ausgeben

// unkorrigierter Takt (SystemCoreClock) zum Vergleich mit 1pps-Signal
        if(gps_aktiv) {
          if(!gps_laeuft) {
            loesche_gps_filter();               // gps war noch nicht aktiviert
            gps_laeuft = 1;                     // ist jetzt initialisiert
          }  
          F_temp = (double)Fref_mess_ereignisse/(double)Fref_mess_zeit * 
                          (SystemCoreClock)/PIO_TIME_DIV;
          if(F_temp >= FREF1_LOW && Fref_ergebnis <= FREF1_HIGH) {     
            Fref_temp = FREF1;                  // 1 Hz
          }   
          else if(F_temp >= FREF2_LOW && F_temp <= FREF2_HIGH) {
            Fref_temp = FREF2;                  // 10 kHz
          } 
          else if(F_temp >= FREF3_LOW && F_temp <= FREF3_HIGH) {
            Fref_temp = FREF3;                  // 1 MHz
          }  
          else if(F_temp >= FREF4_LOW && F_temp <= FREF4_HIGH) {
            Fref_temp = FREF4;                  // 10 MHz
          }  
          else {
            Fref_temp = 0.0;                    // keine gueltige Fref gefunden
            loesche_gps_filter();               // gps war noch nicht aktiviert
          }  
          if(Fref_temp != Fref_aktuell) {       // Fref geaendert
            loesche_gps_filter();               // gps war noch nicht aktiviert
            Fref_aktuell = Fref_temp;
          }  
          if(gps_totzeit) {                     // Einschwingen abwarten
            gps_totzeit--;
          } else {
            F_temp /= Fref_temp;                // auf 1 Hz normiert
            bewerte_gps_signal(F_temp);         // nur engtoleriertes Signal verwenden
          }  
        } else {
          gps_laeuft = 0;                       // kein gueltiges Signal
        }  
      
        eff_zeit2 = (double)(Fref_mess_zeit ) / F_ref;  // Zeit in Sekunden
        Fref_ergebnis = ((double)Fref_mess_ereignisse) / eff_zeit2; // Frequenz berechnen
        Fref_ergebnis += Fref_ergebnis*(F1_offset*F_OFFSET_SCHRITT);
        Fref_ergebnis *= F2_faktor;
        eff_stellen = F2_stellen;
        if(!eff_stellen) eff_stellen = berechne_stellen(eff_zeit2);
        Fx_str_mittig("",Fref_ergebnis, WANDEL_F, s, eff_stellen);
        if(gps_aktiv)
          lcd_zeile3((char *)str0[sprache]); 
        else
          lcd_zeile3((char *)str1[sprache]);  
        lcd_messwert_zeile4(s); 
        if(ser_ausgabe_kanal == F2) {
          wandel_Fx("",Fref_ergebnis, WANDEL_F, s, eff_stellen);
          sende_str1(s);
        }  
      } else 
        Fref_messwert_vorhanden = 1;            // sperre wieder aufheben
    }
  }
}  
